package Agent;

#
# Agent.pm module implements Agent interface
#

use strict;
use warnings;

use XmlNode;
use CommonXmlNodes;

use AgentConfig;
use Dumper;
use Storage::Storage;
use ContentDumperBase;
use DatabaseContentDumper;
use ContentDumper;

use FormatConverter;
use LimitsAndTemplates;
use SpamAssassinCfg;
use EncodeBase64;
use IDN;
use DateUtils;
use CustomLogging;

use Preferences;

#
# Begin interface methods
#

my $do_gzip = undef;
my $agentWorkDir;
my $contentDumperBase = ContentDumperBase->new(Storage::Storage::createFileStorage($do_gzip, getWorkDir()));

sub enableCompression {
  $do_gzip = 1;
}

sub getCompressionStatus {
  return $do_gzip;
}

sub setWorkDir {
  my $workDir = shift;
  $agentWorkDir = $workDir;
}

sub getWorkDir {
  return $agentWorkDir || AgentConfig::cwd();
}

sub getContentTransportDescription{
  return $contentDumperBase->getContentTransportDescription();
}

sub getAgentName {
  return "Confixx";
}

sub getAgentVersion {
  return '10.2';
}

sub getAgentStatus {
  Logging::trace("Getting status...");

  # TODO: check confixx version, pre-requirements, etc
  return XmlNode->new('agent-status');
}

sub getAdminPreferences {
  my $adminPersonalInfo = Dumper::getAdminPersonalInfo();
  my $adminInfo = Dumper::getAdminInfo();
  
  $adminPersonalInfo->{'language'} = $adminInfo->{'language'};

  return XmlNode->new('preferences',
    'children' => [_makePinfoNodes(_convertUserPersonalInfo(undef, $adminPersonalInfo))]
  )
}

# Should return an array of reseller identifiers, that could be a number, name or guid.
# Should be unique through migration dump.
sub getResellers {
  Logging::trace("Getting resellers...");
  return Dumper::getResellers();
}

sub getReseller {
  # reseller identifier from 'getResellers' result
  my $resellerId = shift;

  Logging::trace("Getting reseller dump for '$resellerId' reseller...");

  my $resellerInfo = Dumper::getResellerInfo($resellerId);

  my @templatesNodes = map {
    CommonXmlNodes::domainTemplate($_->{'name'}, $_->{'items'} )
  } LimitsAndTemplates::getClientTemplates($resellerId);

  my $numberOfSystemDomains = scalar(Dumper::getClients($resellerId));    # is equal number of clients

  return XmlNode->new('reseller',
    'attributes' => {
      'id' => $resellerId,
      'name' => $resellerId,
      'guid' => ''
    },
    'children' => [
      XmlNode->new('preferences',
        'children' => [
          _makePinfoNodes(_convertUserPersonalInfo($resellerId, $resellerInfo)),
          @templatesNodes
        ]
      ),
      _makeUserPropertiesNode($resellerInfo->{'longpw'}, !$resellerInfo->{'gesperrt'}),
      CommonXmlNodes::limitsAndPermissions(
        LimitsAndTemplates::getResellerLimits($resellerId, $numberOfSystemDomains),
        LimitsAndTemplates::getResellerPermissions($resellerId)
      ),
      CommonXmlNodes::ipPool(getResellerIPPool($resellerId))
    ]
  );
}

# Should return an array of clients identifiers, that could be a number, name or guid.
# Should be unique through migration dump
sub getClients {
  my $owner = shift;

  unless (Dumper::isResellerId($owner)) {
    # clients are owned by resellers only
    return ();
  }

  Logging::trace("Getting clients list for $owner reseller...");
  
  return Dumper::getClients($owner);
}

sub getClient {
  my ($clientId, $dumpOptions) = @_;

  # client identifier from 'getClients' result,
  Logging::trace("Getting client dump for '$clientId' client...");

  my $clientInfo = Dumper::getClientInfo($clientId);
  
  return XmlNode->new('client',
    'attributes' => {
      'id' => $clientId, 
      'name' => $clientId,
      'guid' => ''
    },
    'children' => [ 
      XmlNode->new('preferences',
        'children' => [_makePinfoNodes(_convertUserPersonalInfo($clientId, $clientInfo))]
      ),
      _makeUserPropertiesNode($clientInfo->{'longpw'}, !$clientInfo->{'gesperrt'}),
      CommonXmlNodes::ipPool(getClientIPPool($clientId)),
      XmlNode->new('domains', 
        'children' => [ _getSystemDomainNode($clientId, $clientInfo) ]
      )
    ]
  );
}

# In current implementation domains are not separated from clients.
#
# Reason: it is not clear how to migrate databases, user's content and other objects
# that are related to a customer, not a domain. The problem rises
# when you try to implement the shallow dump, with the following strategy:
# - if some element should be removed, all its children should go to its parent element
# - if parent element should be removed too, all the children should go to a parent of the parent
# - and so on.
#
# For example user selected a domain, but has not selected a customer. Should the content and databases
# go to a reseller? And if no domains are selected and customer is not selected? Should we dump
# the content? And there are a lot of other such questions, which could not be answered in a simple way.
sub getDomains {
  return ();
}

# <doc>
# == IP migration ==
#
# * Each customer has an IP address - shared or dedicated to a customer. 
# This single IP is used by all customer's domains. 
# Mapping: added to IP pool of all customer's domains. 
# * Each server IP can be:
# ** (only one) standard server IP - used by all resellers (and their customers) that do now own other IPs. Mapping: added to a pool of all resellers without assigned IPs.
# ** assigned to a reseller - used as a shared IP for customers of the reseller. Mapping: added to reseller-owner IP pool as a shared IP, not an exclusive one.
# ''Reason'': bug (problem) in Plesk - this information is stored in 'server' node, which (for some reasons) can only be restored from backup, not migrated.
# ** assigned to a reseller and to a customer - used as a dedicated IP for the customer of the reseller. 
# Mapping: added to reseller-owner IP pool as dedicated IP, customer-owner IP pool as dedicated IP, customer's main domain as dedicated IP.
# ** not assigned non-standard IP - not used by anyone.
#
# </doc>
sub getResellerIPPool {
  my ($reseller) = @_;

  my @exclusiveIPs = Dumper::getResellerExclusiveIPs($reseller);

  if (@exclusiveIPs) {
    my @ips = map { 
      {
        # from the reseller's point of view (every IP can have another type from the admin's point of view)
        'type' => $_->{'kunde'} eq '' ? 'shared' : 'exclusive', 
        'ip' => $_->{'ip'}
      } 
    } @exclusiveIPs;

    return \@ips;
  } else {
    my $resellerInfo = Dumper::getResellerInfo($reseller);
    return [{'type' => 'shared', 'ip' => $resellerInfo->{'standardip'}}];
  }
}

# Client's IP pool always consists of one IP 
sub getClientIPPool {
  my ($client) = @_;

  return [getClientIP($client)];
}

sub getServerNode {
  return XmlNode->new('server', 
    'children' => [
        _makeDnsSettingsNode()
    ]
  );
}

#
# End of interface methods
#

sub _getSystemDomainNode {
  my ($clientName, $clientInfo_ptr) = @_;

  my @domainIds = Dumper::getDomains($clientName);
  my $clientSystemDomain = Dumper::getClientSystemDomain($clientName);
  my $numberOfSites = scalar(@domainIds) + 1;  # "+1" required for system domain

  Logging::trace("Getting dump of a system domain '$clientSystemDomain' of a '$clientName' client");

  return XmlNode->new('domain',
    'attributes' => {
      'name' => $clientSystemDomain,
      'guid' => '',
      'www' => 'false'
    },
    'children' => [
      XmlNode->new('preferences'),
      _getSystemDomainPropertiesNode($clientName),
      CommonXmlNodes::limitsAndPermissions(
        LimitsAndTemplates::getClientLimits($clientName, $numberOfSites),
        LimitsAndTemplates::getClientPermissions($clientName)
      ),
      _getMailSystem($clientSystemDomain, $clientInfo_ptr, 1),
      _makeDatabaseListNode($clientName, $clientSystemDomain), # clientSystemDomain is used solely to name dump file
      _makeDomainMaillistsNode($clientInfo_ptr),
      CommonXmlNodes::traffic(getTraffic($clientInfo_ptr->{'kunde'})),
      _makeSslNode($clientInfo_ptr),
      _makeDomainHostingNode($clientInfo_ptr, $clientSystemDomain, \@domainIds)
    ]
  );
}

sub _getSystemDomainPropertiesNode {
  my ($clientName) = @_;

  Logging::trace("Making properties for system domain of client '$clientName' ... ");

  my $clientInfo = Dumper::getClientInfo($clientName);

  my $clientIP = getClientIP($clientName);

  return XmlNode->new('properties', 
    'children' => [
      CommonXmlNodes::ip($clientIP->{'ip'}, $clientIP->{'type'}),
      CommonXmlNodes::status(!$clientInfo->{'gesperrt'}, 'admin')
    ]
  );
}

#
# create <mailsystem> node
#
sub _getMailSystem {
  my ($domainName, $clientInfo, $isSystemDomain) = @_;
  my $client = $clientInfo->{'kunde'};

  Logging::trace("Getting emails for domain '$domainName' ...");

  my ($mailUsers, $catchAll) = _mappingMailUsers($domainName, $client, $isSystemDomain);

  if (!%{$mailUsers} && !defined($catchAll)) {
    return undef;
  }

  my $mailSystemNode = XmlNode->new('mailsystem',
    'children' => [
        XmlNode->new('properties', 'children' => [
            CommonXmlNodes::status(1, 'admin')
        ])
    ]
  );

  #
  # add <mailusers> according %$mailUsers
  #
  if (%{$mailUsers}) {
    my $mailUsersNode = XmlNode->new('mailusers');
    while (my ($emailPrefix, $emailSettings) = each %{$mailUsers}) {
        $mailUsersNode->addChild(_getMailUserNode($domainName, $clientInfo, $emailPrefix, $emailSettings));
        _logEmailMapping($emailPrefix, $domainName, $emailSettings);
    }

    $mailSystemNode->addChild($mailUsersNode);
  }

  my $catchAllNode = XmlNode->new('catch-all');
  $catchAllNode->setText(
    defined($catchAll) ? _convertEmailAddressToUtf8($catchAll) : "reject",
    'UTF-8'
  );

  $mailSystemNode->addChild(XmlNode->new('preferences',
    'children' => [
      $catchAllNode,
      XmlNode->new('web-mail', 'content' => 'horde'),
    ]
  ));


  return $mailSystemNode;
}

#
# return hash of <mailusers> nodes and catch-all setting
# TODO: explain email mapping here
#
sub _mappingMailUsers {
  my ($domainName, $client, $isSystemDomain) = @_;

  my %mailUsers;        # table for creating mailuser's xml nodes. Format this table is:
    # $mailUsers{'email'}                    - E-mail Address in Plesk (mailuser node in dump.xml)
    # $mailUsers{'email'}->{'mbox'}          - name of POP3 box in Confixx ('email' can be without mbox)
    # @{$mailUsers{'email'}->{'alias'}}      - ARRAY of E-mail Address aliases
    # @{$mailUsers{'email'}->{'forwarding'}} - ARRAY of E-mail Address forwarding
    # $mailUsers{'email'}->{'isspecialemail'}  - indicate, that this mailuser is specially generated (for logging in _logEmailMapping)

  #
  # In special system(fake) domain:
  # create special emails(mailusers) for mboxes without emails
  #
  if ($isSystemDomain) {
      map { $mailUsers{$_}->{'mbox'} = $_; $mailUsers{$_}->{'isspecialemail'} = 1; } Dumper::getMboxesWithoutEmails($client);

      return \%mailUsers;
  }

  my $catchAll;
  my $mboxToEmailMap = Dumper::getMboxToEmailMap($client);

  #
  # first loop: fill %mailUsers from email table for this domain
  #             main mapping actions(selections) here
  #
  foreach my $emailForward (Dumper::getEmailPrefixToDestinationTable($domainName)) {
      my $emailPrefix = $emailForward->{'prefix'};
      my $destination = $emailForward->{'pop3'};

      if ($destination =~ m/@/) {
        # create forward
        push @{$mailUsers{$emailPrefix}->{'forwarding'}}, $destination;
      } else {  # $destination is mbox here
        if ("$emailPrefix\@$domainName" ne $mboxToEmailMap->{$destination}) {
          # if email is not main/first email for this $destination mbox - create forward to main email
          push @{$mailUsers{$emailPrefix}->{'forwarding'}}, $mboxToEmailMap->{$destination};
        } else {
          if (defined($mailUsers{$emailPrefix}->{'mbox'})) {
              # Find more than one mbox for this email. So, create new Plesk E-Mail Address for this mbox
              my $newEmailPrefix = _generateEmailPrefixName($destination, \%mailUsers, $domainName);
              $mailUsers{$newEmailPrefix}->{'mbox'} = $destination;

              push @{$mailUsers{$emailPrefix}->{'forwarding'}}, "$newEmailPrefix\@$domainName";
          } else {
              # define Confixx POP3 mbox for this email
              $mailUsers{$emailPrefix}->{'mbox'} = $destination;
          }
        }
      }

      if (!defined($mailUsers{$emailPrefix}->{'autoresponder'})
        && (my $autoresponder = Dumper::getEmailAutoresponder($emailPrefix, $domainName))) {
          $mailUsers{$emailPrefix}->{'autoresponder'} = $autoresponder;
      }
  }

  #
  # second loop: get destinations from "*" email prefixes from ForwardTo Confixx
  #              and migrate it in Plesk catch-all settings
  #
  if (my @catchAllDestinations = Dumper::getCatchAllDestinations($domainName)) {
    if (@catchAllDestinations == 1) {
      # simple case: one "*" - one catch-all
      $catchAll = _checkFullCatchAllDestination($catchAllDestinations[0], $domainName, $mboxToEmailMap, \%mailUsers);
    } else {
      # Plesk can have only ONE catch-all email, so we create special email for all catch-all emails
      my $catchAllSpecialMailUser = _generateEmailPrefixName("catch-all", \%mailUsers, $domainName);
      $catchAll = "$catchAllSpecialMailUser\@$domainName";

      Logging::trace("Created $catchAll email for all catch-all emails of this domain.");

      foreach my $destination (@catchAllDestinations) {
        push @{$mailUsers{$catchAllSpecialMailUser}->{'forwarding'}},
            _checkFullCatchAllDestination($destination, $domainName, $mboxToEmailMap, \%mailUsers);
      }
    }
    if (Dumper::getEmailAutoresponder("*", $domainName)) {
      # Autoresponders for *@domain.com emails (catch-all) not migrated, because:
      # - not supported by Plesk
      # - not migrated by old migrator
      # - Very bad idea to create autoresponder for nonexistens users of domain (it will be generate tons of spam!)
      Logging::warning("Autoresponder for *\@$domainName email (catch-all) not migrated, because not supported.");
    }
  }

  #
  # bug #68368 "full mesh mailmaping with two email and two mailbox"
  # next loop for resolving it: remove duplicate 'forwarding' entries
  #
  while (my ($emailPrefix, $emailSettings) = each %mailUsers) {
    next unless(defined($emailSettings->{'forwarding'}));

    my %uniqueEmailForwardings = map { $_ => 1 } @{$emailSettings->{'forwarding'}};
    @{$emailSettings->{'forwarding'}} = keys %uniqueEmailForwardings;
  }

  #
  # next loop:
  #     Create aliases (from $mailUsers{}->{'forwardings'})
  # collapse emails forwardings into aliases by the following conditions:
  # - this email has only one <forwarding>
  # - this <forwarding> contains same $domainName
  # - this <forwarding> pointer to existing mailuser
  # - this email has not mbox
  # - this email has not autoresponders
  #
  if (%mailUsers) {
    while (my ($emailPrefix, $emailSettings) = each %mailUsers) {
        if (defined($emailSettings->{'forwarding'})                                 # check exist forwarding
          && !defined($emailSettings->{'mbox'})                                     # check not exist mbox
          && !defined($emailSettings->{'autoresponder'})                            # check not exist autoresponder
          && scalar(@{$emailSettings->{'forwarding'}}) == 1                         # check one forwarding
          && @{$emailSettings->{'forwarding'}}[0] =~ m/^([^@]+)\@$domainName$/) {   # check same domain
            my $alias = $1;     # get prefix of this forwarding email

            if (!defined($mailUsers{$alias})) {                                     # check exist mailUsers
                next;
            }
            push @{$mailUsers{$alias}->{'alias'}}, $emailPrefix;                    # create alias
            delete($mailUsers{$emailPrefix});                                       # remove this mailuser with simple forwarding
        }
    }
  }

  return (\%mailUsers, $catchAll);
}

sub _checkFullCatchAllDestination {
  my ($catchDestination, $domainName, $mboxToEmailMap, $mailUsers) = @_;

  # $catchDestination is email
  if ($catchDestination =~ m/@/) {
    return $catchDestination;
  }

  # $catchDestination is mbox. Look for existing email for this mbox
  if (defined($mboxToEmailMap->{$catchDestination})) {
    return $mboxToEmailMap->{$catchDestination}
  }

  # this mbox has no main email. Create special email for this mbox
  my $newEmailPrefix = _generateEmailPrefixName($catchDestination, $mailUsers, $domainName);
  $mailUsers->{$newEmailPrefix}->{'mbox'} = $catchDestination;
  return "$newEmailPrefix\@$domainName";
}

# create unique EmailPrefix for this Confixx mbox
sub _generateEmailPrefixName {
  my ($mboxname, $mailUsers, $domainName) = @_;
  use constant MAIL_USERNAME_LENGTH_LIMIT => 99;
  use constant MAIL_USERNAME_INDEX_LENGTH => 4;
  my $newMboxname = $mboxname;
  my $index = 0;

  while (defined($mailUsers->{$newMboxname}) || Dumper::checkExistEmail($newMboxname, $domainName)) {
    $newMboxname = _getStringIndexWithSuffix("${mboxname}_mb", ++$index, MAIL_USERNAME_LENGTH_LIMIT, MAIL_USERNAME_INDEX_LENGTH);
  }

  $mailUsers->{$newMboxname}->{'isspecialemail'} = 1;

  return $newMboxname;
}

sub _getStringIndexWithSuffix {
  my ($string, $valueIndex, $maxLength, $indexStrLength) = @_;

  return undef if ($valueIndex >= 10 ** $indexStrLength);

  my $strIndex = sprintf("%0${indexStrLength}d", $valueIndex);
  return (substr($string, 0, $maxLength - $indexStrLength) . $strIndex);
}

sub _getMailUserNode {
  my ($domainName, $clientInfo, $emailPrefix, $emailSettings) = @_;
  my $mailBoxNode;
  my $mboxName = $emailSettings->{'mbox'};
  my %attributes = (
    'name' => $emailPrefix,
    'forwarding-enabled' => (defined($emailSettings->{'forwarding'}) && @{$emailSettings->{'forwarding'}}) ? 'true' : 'false'
  );

  Logging::debug("Create mailuser '$emailPrefix' node for '$domainName' domain/site");

  my $password;
  if (defined($mboxName)) {
    my $contentNode = ContentDumper::getMailBoxContent($contentDumperBase, $mboxName, $domainName);

    $mailBoxNode = XmlNode->new('mailbox',
      'attributes' => {
        'type'      => 'mdir',
        'enabled'   => 'true'
      },
      'children'   => [
        $contentNode
      ]
    );

    my $mboxInfo = Dumper::getMboxInfo($mboxName);
    $password = $mboxInfo->{'longpw'};
    $password =~ s/^!//;            # remove password lock

    $attributes{'mailbox-quota'} = _getMboxQuota($mboxInfo->{'maxkbhard'}, $mboxName);
  }

  my @aliasNodes = map { XmlNode->new('alias', 'content' => $_) } @{$emailSettings->{'alias'}};
  my @forwardingNodes = map { 
    my $forwardingNode = XmlNode->new('forwarding');
    $forwardingNode->setText(_convertEmailAddressToUtf8($_), 'UTF-8'); 
    $forwardingNode;
  } @{$emailSettings->{'forwarding'}};

  return XmlNode->new('mailuser',
    'attributes' => \%attributes,
    'children' => [
      XmlNode->new('properties', 'children' => [
        defined($password) ? CommonXmlNodes::encryptedPassword($password) : CommonXmlNodes::emptyPassword()
      ]),
      XmlNode->new('preferences', 'children' => [
        $mailBoxNode,
        @aliasNodes,
        @forwardingNodes,
        _getEmailAutoresponders($emailSettings->{'autoresponder'}),
        defined($mboxName) ? _getEmailAddressBook($mboxName) : undef,
        defined($mboxName) && ConfixxConfig::getValue('spamassassin_support') ? 
          CommonXmlNodes::spamassassin(_getSpamassassinOptions($mboxName, $clientInfo)) : ()
      ])
    ]
  );
}

# Convert e-mail address from latin1 and punycode to UTF-8
sub _convertEmailAddressToUtf8 {
  my ($emailAddress) = @_;

  # we use an ungreedy match to match the last '@' symbol 
  # as according to RFC 822 there could be a '@' in a local-part of e-mail address,
  # but '@' is not allowed in domain name
  # this could be also written as /^(.*)@([^@]*)$/ if it is more clear for you
  if ($emailAddress =~ /^(.*)@(.*?)$/s) { 
    # local part is called $emailPrefix here
    my ($emailPrefix, $domain) = ($1, $2);
    my ($canConvertIdnDomain, $reason) = IDN::canConvert();
    if ($canConvertIdnDomain && IDN::isIdnDomain($domain)) {
      return Encoding::encode($emailPrefix) . '@' . IDN::idnToUtf8($domain);
    } else {
      return Encoding::encode($emailAddress);
    }
  } else {
    return Encoding::encode($emailAddress);
  }
}

sub _getMboxQuota {
  my ($mboxQuota, $mboxName) = @_;

  my $adminInfo = Dumper::getAdminInfo();
  my $globalMboxQuota = LimitsAndTemplates::convertBytesLimitValue($adminInfo->{'popmaxkbhard'});
  $mboxQuota = LimitsAndTemplates::convertBytesLimitValue($mboxQuota);

  if (!ConfixxConfig::getValue('mail_quota')) {
    return LimitsAndTemplates::PLESK_RESOURCE_UNLIMITED;
  }

  if ($mboxQuota == LimitsAndTemplates::PLESK_RESOURCE_UNLIMITED || $mboxQuota == 0) { # sometimes, in Confixx DB can be zero values for mbox quotas
    return $globalMboxQuota;
  }

  if ($globalMboxQuota != LimitsAndTemplates::PLESK_RESOURCE_UNLIMITED && $mboxQuota > $globalMboxQuota) {
    Logging::warning("Quota $mboxQuota of mailbox '$mboxName' is more than global mailbox limit $globalMboxQuota. Setting mailbox quota value to global quota value.");
    $mboxQuota = $globalMboxQuota;
  }

  return $mboxQuota;
}

sub _getEmailAutoresponders {
  my ($autoresponder) = @_;

  if (!defined($autoresponder)) {
    return undef;
  }

  my $text      = defined($autoresponder->{'text'})     ? $autoresponder->{'text'} : "";
  my $fromName  = defined($autoresponder->{'fromname'}) ? $autoresponder->{'fromname'} : "";
  my $fromEmail = defined($autoresponder->{'fromemail'})? $autoresponder->{'fromemail'} : "";
  my $subject   = defined($autoresponder->{'subject'})  ? $autoresponder->{'subject'} : "";

  if ( $fromName ne "" || $fromEmail ne "") {
    $text .= "\n--\n$fromName";
    if ($fromEmail ne "") {
        $text .= " <$fromEmail>";
    }
  }

  return XmlNode->new('autoresponders',
    'children' => [
      XmlNode->new('autoresponder',
        'attributes' => {
          'content-type' => 'text/plain',
          'subject' => EncodeBase64::encode($subject),
          'status' => 'on'
        },
        'children' => [
          XmlNode->new('text',
            'content' => EncodeBase64::encode($text))
        ]
      )
    ]
  );
}

sub _getEmailAddressBook {
  my ($mboxName) = @_;
  my @addressBook = Dumper::getEmailAddressBook($mboxName);

  unless (@addressBook) {
    return undef;
  }


  # TODO: make correct attributes
  return XmlNode->new('addressbook',
    'children' => [
      map {
        XmlNode->new('addressbook-contact',
          'attributes' => {
            'alias'         => EncodeBase64::encode($_->{'name'}),                   # We can not use 'name' attribute for it, because we do not know format of this attribute.
            'email'         => EncodeBase64::encode($_->{'email'}),
            'fax'           => EncodeBase64::encode($_->{'tel_fax'}),
            'home-phone'    => EncodeBase64::encode($_->{'tel_home'}),
            'id'            => "some_required_string_${mboxName}_" . $_->{'ident'} , # It is useless string, but must be unique for every record for work.
            'mobile-phone'  => EncodeBase64::encode($_->{'tel_mobile'}),
            'notes'         => EncodeBase64::encode("Address: " . $_->{'address'}),  # We can not use 'work-address' for it, because we do not know format of this attribute.
            'work-phone'    => EncodeBase64::encode($_->{'tel_work'})
          }
        )
      } @addressBook
    ]
  );
}

use constant MAIL_MAPPING_LOG_FILE => 'Confixx-mail-mapping.txt';

sub _logEmailMapping {
  my ($emailPrefix, $domainName, $emailSettings) = @_;

  my $domainNameUtf8 = _convertEmailAddressToUtf8($domainName);
  my $confixxEmailUtf8 = _convertEmailAddressToUtf8("$emailPrefix\@$domainName");
  my $mboxName = $emailSettings->{'mbox'};
  my $writeToLog = sub { CustomLogging::addToLog(MAIL_MAPPING_LOG_FILE, $_[0] . "\n") };

  if (defined($emailSettings->{'isspecialemail'})) {
    if (defined($mboxName)) {
      $writeToLog->("Confixx's POP3 box '$mboxName' is migrated into Plesk's e-mail address '$confixxEmailUtf8'.");
    } else {
      $writeToLog->("Confixx's '*' e-mail addresses are migrated into Plesk's e-mail address '$confixxEmailUtf8'.");
    }
  } else {
    my $mboxDescription = defined($mboxName) ? " with mailbox '$mboxName'" : "";

    $writeToLog->("Confixx's e-mail address '$confixxEmailUtf8' is migrated as is$mboxDescription.");
  }

  my @forwardingsUtf8 = map { _convertEmailAddressToUtf8($_); } @{$emailSettings->{'forwarding'}};
  foreach my $forwarding (@forwardingsUtf8) {
    $writeToLog->("Additionally, e-mail address '$forwarding' is added to forwardings of Plesk's e-mail address '$confixxEmailUtf8'.");
  }

  my @aliasesUtf8 = map { _convertEmailAddressToUtf8($_); } @{$emailSettings->{'alias'}};
  foreach my $alias (@aliasesUtf8) {
    $writeToLog->("Confixx's e-mail address '$alias\@$domainNameUtf8' is migrated into an alias of Plesk's e-mail address '$confixxEmailUtf8'.");
  }
}

# <doc>
# = Mail =
# == Spamassassin ==
# Custom spamassassin settings (that are edited in "For advanced users" mode) are not migrated. 
# ''Reason'': there is no such feature in Plesk.
#
# Action is always 'mark', other actions ('delete' and 'move') are not migrated because action is set
# outside of spamassassin config, and Confixx has no control on this: it's difficult to determine 
# which of them is enabled as it requires at least parsing procmailrc. Moreover it is not possible
# to determine if procmailrc is used, or if there is another solution.
#
# Server-wide spamassassin settings are not migrated due to PMM limitations. However if 'hits' or 
# 'subj-text' are not set in user spam preferences, they are taken from spamassassin global configuration.
#
# </doc>

use constant SPAMASSASSIN_ACTION_MARK => 'mark';

use constant SPAMASSASSIN_MAPPING_LOG_FILE => 'Confixx-spamassassin-mapping.txt';

sub _getSpamassassinOptions {
  my ($mailbox, $clientInfo) = @_;
  my %options;
  
  my @spamPreferences = Dumper::getMboxSpamPreferences($mailbox);

  $options{'status'} = $clientInfo->{'spamfilter'};
  $options{'action'} = SPAMASSASSIN_ACTION_MARK;
  $options{'blacklist'} = [];
  $options{'whilelist'} = [];

  foreach my $preference (@spamPreferences) {
    my ($name, $value) = ($preference->{'preference'}, $preference->{'value'});

    if ($name eq 'rewrite_header') {
      if (defined($options{'subj-text'})) {
        Logging::warning("Duplicate rewrite_header value '$value', using the last one");
      }
      $options{'subj-text'} = $value;
    } elsif ($name eq 'required_score') {
      if (defined($options{'hits'})) {
        Logging::warning("Duplicate required_score value '$value', using the last one");
      }
      $options{'hits'} = $value;
    } elsif ($name eq 'blacklist_from') {
      push(@{$options{'blacklist'}}, $value);
    } elsif ($name eq 'whitelist_from') {
      push(@{$options{'whitelist'}}, $value);
    } else {
      CustomLogging::addToLog(SPAMASSASSIN_MAPPING_LOG_FILE, "Mailbox '$mailbox': SpamAssassin option '$name' with value '$value' was not migrated as there is no equivalent option in Plesk.\n");
    }
  }

  # use default system-wide options if user options are not set
  my %serverWideOptions = _getServerWideSpamassassinOptions();

  if (!defined($options{'hits'}) && defined($serverWideOptions{'hits'})) {
    $options{'hits'} = $serverWideOptions{'hits'};
  }
  
  if (!defined($options{'subj-text'}) && defined($serverWideOptions{'subj-text'})) {
    $options{'subj-text'} = $serverWideOptions{'subj-text'};
  }

  return %options;
}

sub _getServerWideSpamassassinOptions {
  my %result;

  my $config = SpamAssassinCfg::parseConfig(
    ConfixxConfig::getValue('spamassassinConfig'),
    undef # we do not pass home dir as it is not clear what is the home directory
  ); 

  my $hits = SpamAssassinCfg::getConfigRequireScore($config);
  $result{'hits'} = $hits if (defined $hits);

  my $rewriteHeaderArray = SpamAssassinCfg::getConfigRewriteHeadr($config);
  my $rewriteHeaderText;
  if (defined($rewriteHeaderArray) && @{$rewriteHeaderArray} == 2) {
    my ($argumentRewrite, $textRewrite) = @{$rewriteHeaderArray};

    $result{'subj-text'} = $textRewrite if ($argumentRewrite =~ m/subject/i);
  }

  return %result;
}

sub getClientIP {
  my ($client) = @_;

  my $clientInfo = Dumper::getClientInfo($client);
  my $clientIP = $clientInfo->{'ip'};
  my @dedicatedIPs = Dumper::getClientExclusiveIPs($client); 

  if (@dedicatedIPs == 1 && $dedicatedIPs[0] eq $clientIP) {
    return {'type' => 'exclusive', 'ip' => $clientIP};
  } else {
    if (@dedicatedIPs == 1 && $dedicatedIPs[0] ne $clientIP) { # fall back to shared IP
      Logging::warning("Expected that the only exclusive IP is the same as IP from client properties");
    } elsif (@dedicatedIPs > 1) { # fall back to shared IP
      Logging::warning("Expected that every client has 0 or 1 exclusive IPs, but client $client has " . scalar(@dedicatedIPs) . " IPs");
    }
    return {'type' => 'shared', 'ip' => $clientIP };
  }
}


# <doc>
# = Personal information =
# Gender, Customer ID and all definable fields are not mapped.
# ''Reason'': all non-mapped information could be mapped to comment field according to XSD,
# but there is no place in UI where this comment is shown or could be edited.
# 
# For all fields except name, language and locale, no convesion is made.
#
# Full country name should be used as a country field value. Otherwise it will not be migrated. ''Reason'': Plesk supports only a set of fixed values for country field.
#
# All the other fields (company name, phone, fax, address, city, zip code, e-mail) must 
# conform to Plesk format. Otherwise they will not be migrated and Plesk will issue a warning (however migration completes).
# </doc>
sub _convertUserPersonalInfo {
  my ($userId, $userInfo) = @_;

  my %result;

  my $name = _convertUserName($userId, $userInfo->{'firstname'}, $userInfo->{'name'});
  if (defined($name)) {
    $result{'name'} = $name;
  }

  my %usersMap = (
    'firma' => 'company',
    'telefon'=> 'phone',
    'fax' => 'fax',
    'anschrift' => 'address',
    'plzort' => 'city',
    'plz' => 'zip',
    'emailadresse' => 'email' 
  );
  
  while (my ($confixxKey, $pleskKey) = each %usersMap) {
    if ($userInfo->{$confixxKey} ne '') {
      $result{$pleskKey} = $userInfo->{$confixxKey};
    }
  }

  if ($userInfo->{'land'} ne '') {
    my $countryCode = FormatConverter::convertCountry($userInfo->{'land'});
    if (defined($countryCode)) {
      $result{'country'} = $countryCode;
    } else {
      Logging::warning("Failed to convert country code '" . $userInfo->{'land'} . "' for $userId");
    }
  }

  my $locale = FormatConverter::convertLocale($userInfo->{'language'});

  if (defined($locale)) {
    $result{'locale'} = $locale;
  } else {
    Logging::warning("Failed to convert locale '" . $userInfo->{'language'} . "' for $userId");
  }

  return \%result;
}

# <doc>
# == First and last names ==
# Plesk name is a combination of a first name and a last name, if some of them is set.
# If both the first name and the last name are not set - use user id (something like "res5", or "web8") as a name.
# </doc>
sub _convertUserName {
  my ($userId, $firstName, $lastName) = @_;

  if ($firstName ne '' && $lastName ne '') { 
    my @nameComponents;
    push @nameComponents, $firstName if ($firstName ne '');
    push @nameComponents, $lastName if ($lastName ne '');
    return join ' ', @nameComponents;
  } else {
    return $userId;
  }
}

# Users CP password and enabled/disabled status is migrated
sub _makeUserPropertiesNode {
  my ($password, $isEnabled) = @_;
  return XmlNode->new('properties', 
    'children' => [
      CommonXmlNodes::encryptedPassword($password),
      CommonXmlNodes::status($isEnabled, 'admin')
    ]
  );
}

# TODO migrator to Plesk9 also does the following questionable things:
# - add a webmail record manually: makeDnsRecord($dnsZoneNode, 'A', "webmail.$src", $dst);
# - read NS record values from reseller's or admin's settings instead of zone file
sub _makeDomainDnsZoneNode {
  my ( $domainInfo ) = @_;
  my $domainName = $domainInfo->{'domain'};
  Logging::trace("DNS status for domain $domainName: ".$domainInfo->{'dns'});

  # generate admin e-mail for domain the same way Confixx does it
  my $reseller  = Dumper::getResellerInfo($domainInfo->{'anbieter'});
  my $emailHost = $reseller->{'pns'};
  # take the domain part of the name: ns.example.com -> example.com
  $emailHost    =~ s/^[^\.]+\.//;
  my $email     = 'hostmaster.'.('' ne $emailHost ? $emailHost : $domainName);

  my @rawDnsRecords = Dumper::parseDnsRecords(Dumper::getDnsRecords($domainName));
  my $status     = $domainInfo->{'dns'};

  return _makeDnsZoneNode($domainName, $email, $status, \@rawDnsRecords);
}
  
sub _makeDnsZoneNode {
  my ($domainName, $email, $status, $dnsRecords) = @_;

  # create a SOA record with the same (hard-coded in confixx_updatescript.pl) parameters as Confixx does
  my @soaRecord = (
      "$domainName.",       # SRC
      86400,                # TTL
      'IN',                 # CLASS
      'SOA',                # TYPE
      '',                   # NS (not used by Plesk migrator)
      '',                   # admin (not used by Plesk migrator)
      '',                   # serial (not used by Plesk migrator)
      10800,                # refresh
      3600,                 # retry
      604800,               # expire
      86400                 # minimum
    );

  XmlNode->new('dns-zone', (
    'attributes' => {
        'email'         => $email,
        'serial-format' => 'YYYYMMDDNN',
        'type'          => 'master',
    },
    'children' => [
        CommonXmlNodes::status($status, 'admin'),
        CommonXmlNodes::dnsZoneParams(@soaRecord),
        # note that DNS records will be found only if DNS is enabled for domain
        map { CommonXmlNodes::dnsRecord(@$_) } @$dnsRecords
    ] 
  ));
}

sub _makeDomainHostingNode {
  my ($customerInfo_ptr, $domainName, $domainIds) = @_;

  my %pDirs = Dumper::getProtectedDirectories($customerInfo_ptr, '/');

  my $pHostingNode = XmlNode->new('phosting',
    'attributes'  => {
      'webstat' => ($customerInfo_ptr->{'statistik'} == 1 ? 'webalizer' :
                    $customerInfo_ptr->{'awstats'} == 1 ? 'awstats' : 'false'),
      'errdocs' => $customerInfo_ptr->{'fehlerseiten'} ? 'true' : 'false',
      'wu_script' => 'true',
      'guid' => '',
      'owner-guid' => '',
      'www-root' => 'httpdocs',
      'cgi_bin_mode' => 'www-root'
    },
    'children' => [
      ContentDumper::getPHostingContent($contentDumperBase, $customerInfo_ptr->{'kunde'}, $domainName, '/', \%pDirs),
      _makeHostingPreferencesNode($customerInfo_ptr, $domainName, \%pDirs),
      XmlNode->new('limits-and-permissions', 'children' => [ _makeScriptingNode($customerInfo_ptr) ]),
    ]
  );

  my @ftpUserNodes = _makeFtpUserNodes($customerInfo_ptr->{'kunde'});
  if (@ftpUserNodes) {
    $pHostingNode->addChild(XmlNode->new('ftpusers', 'children' => [ @ftpUserNodes ] ));
  }

  my @siteNodes = _getDomainSites($customerInfo_ptr, $domainIds);
  if (@siteNodes) {
    $pHostingNode->addChild(XmlNode->new('sites', 'children' => [ @siteNodes ] ));
  }

  return $pHostingNode;
}

sub _makeFtpUserNodes {
  my ($clientName) = @_;

  return map {
    XmlNode->new('ftpuser',
      'attributes' => {
        'name' => $_->{'account'}
      },
      'children' => [
        CommonXmlNodes::sysuser($_->{'account'}, $_->{'longpw'}, 'httpdocs/' . $_->{'pfad'}),
        XmlNode->new('permission', 'content' => 'upload'),
        XmlNode->new('permission', 'content' => 'download')
      ]
    );
  } Dumper::getFtpRecords($clientName);
}

sub _getDomainSites {
  my ($clientInfo_ptr, $siteNames) = @_;

  my @result;
  for my $siteName (@$siteNames) {
    my $siteInfo_ptr = Dumper::getDomainInfo($siteName);
    my $siteNode = XmlNode->new('site',
      'attributes' => {
        'guid' => '',
        'www' => 'false'
      },
      'children' => [
      	XmlNode->new('preferences'), # there are no own preferences, but this node is required by schema
        XmlNode->new('properties', 
          'children' => [
            CommonXmlNodes::status(!$clientInfo_ptr->{'gesperrt'}, 'admin'),
            _makeDomainDnsZoneNode($siteInfo_ptr)
          ]
        ),
        _getMailSystem($siteName, $clientInfo_ptr, 0),
        _makeSiteMaillistsNode($clientInfo_ptr, $siteName),
        _makeSiteHostingNode($clientInfo_ptr, $siteInfo_ptr)
      ]
    );

    my ($canConvertIdnDomain, $reason) = IDN::canConvert();
    if (IDN::isIdnDomain($siteName) && !$canConvertIdnDomain) {
      Logging::warning(
        "Can not convert IDN domain $siteName to UTF-8. Reason: $reason. " . 
        "You will have this domain in punycode after migration, however Plesk supports UTF-8 domains."
      );
    }

    $siteNode->setAttribute('name', IDN::idnToUtf8($siteName), 'UTF-8'); # domain name in UTF-8, for resulting XML 
    push @result, $siteNode;
  }

  return @result;
}

sub _makeSiteHostingNode {
  my ($customerInfo_ptr, $domainInfo_ptr) = @_;

  if ($domainInfo_ptr->{'pfad'} =~ /^http(s)?:/) {
    return XmlNode->new('shosting',
      'children' => [
        XmlNode->new('url', 'content' => $domainInfo_ptr->{'pfad'})
      ]
    );
  }

  my $domainName = $domainInfo_ptr->{'domain'};

  my $vhostFileContent = _getHttpdSpecialsVhostConfigurationText($domainName);

  # not HTTPD specials, but a nice redirect added into the same vhostFileContent to migrate Confixx's http->https feature
  use constant CSSL_NONE => 0;
  use constant CSSL_HTTPS => 1;
  use constant CSSL_HTTPS_PLUS_REDIRECT => 2;
  if ($domainInfo_ptr->{'cssl'} == CSSL_HTTPS_PLUS_REDIRECT) {
    $vhostFileContent .= "\nRedirect / https://" . $domainName . "/\n";
  }
  Logging::debug("vhost.conf content: \n$vhostFileContent");

  my $pHostingNode = XmlNode->new('phosting',
    'attributes'  => {
      'www-root' => 'httpdocs' . $domainInfo_ptr->{'pfad'},
      'https' => $domainInfo_ptr->{'cssl'} == CSSL_NONE  ? 'false' : 'true',
      'cgi_bin_mode' => 'www-root'
    },
    'children' => [
      ContentDumper::getSitePHostingContent($contentDumperBase, $domainName, $vhostFileContent),
      XmlNode->new('preferences'), # there are no own preferences, but this node is required by schema
      XmlNode->new('limits-and-permissions', 'children' => [ _makeScriptingNode($customerInfo_ptr) ])
    ]
  );

  return $pHostingNode;
}

# <doc>
# == HTTPD specials ==
# HTTPD specials are migrated into vhost.conf. Plesk doesn't provide the functionality that
# exists in Confixx: there is no UI for editing HTTPD settings, reseller can't edit HTTPD settings. 
# However existing HTTPD settings are migrated to vhost.conf, without substitution of variables and
# commented by default (see Preferences.pm for more details).
# </doc>
sub _getHttpdSpecialsVhostConfigurationText {
  my ($domainName) = @_;

  my @httpdSpecials = Dumper::getDomainHttpdSpecials($domainName);

  if ($Preferences::commentHttpdSpecials) {
    @httpdSpecials = map { 
      my $option = $_;
      join "\n", 
        map { "# $_"} 
          split /\n/, $option;
    } @httpdSpecials;
  }

  unless (@httpdSpecials) {
    return '';
  }

  return join "\n", ( 
    '# BEGIN HTTPD special options from Confixx',
    '',
    @httpdSpecials,
    '', 
    '# END HTTPD special options from Confixx'
  ); 
}


sub _makeHostingPreferencesNode {
  my ($customerInfo_ptr, $domainName, $pDirs_ptr) = @_;

  my $sysUserNode = CommonXmlNodes::sysuser($customerInfo_ptr->{'kunde'}, $customerInfo_ptr->{'longpw'}, "/var/www/vhosts/". $domainName);
  $sysUserNode->setAttribute('quota', $customerInfo_ptr->{'maxkb'} * 1024) if ($customerInfo_ptr->{'maxkb'} > 0);
  $sysUserNode->setAttribute('shell', $customerInfo_ptr->{'shell'}) if ($customerInfo_ptr->{'shell'});

  return
    XmlNode->new('preferences',
      'children' => [
        $sysUserNode,
        @{_makePDirNodes($pDirs_ptr)},
      ]
    );
}

sub _makeScriptingNode {
  my ($scriptingInfo_ptr) = @_;

  my $scriptingNode = XmlNode->new('scripting',
    'attributes' => {
      'ssi' => $scriptingInfo_ptr->{'ssi'} ? 'true' : 'false',
      'php' => $scriptingInfo_ptr->{'php'} ? 'true' : 'false',
      'cgi' => $scriptingInfo_ptr->{'perl'} ? 'true' : 'false',
      'perl' => $scriptingInfo_ptr->{'perl'} ? 'true' : 'false'
    }
  );

  return $scriptingNode;
}

sub _makeSslNode {
  my ($customerInfo_ptr) = @_;

  my @csslRecords = Dumper::getClientSSLRecords($customerInfo_ptr->{'kunde'});
  unless (@csslRecords) {
    return undef;
  }
  my $csslInfo_ptr = $csslRecords[0];
  unless ($csslInfo_ptr->{'crt'} && $csslInfo_ptr->{'privatekey'}) {
    return undef;
  }

  my $certificatesNode = XmlNode->new('certificates',
    'children' => [
      XmlNode->new('certificate',
        'attributes' => {
          'default' => 'true'
        },
        'children' => [
          XmlNode->new('certificate-data', 'content' => $csslInfo_ptr->{'crt'}),
          XmlNode->new('private-key', 'content' => $csslInfo_ptr->{'privatekey'})
        ]
      )
    ]
  );

  return $certificatesNode;
}

sub _makePDirNodes {
  my ($pDirs_ptr) = @_;

  my @result;
  while (my ($path, $properties) = each %{$pDirs_ptr}) {
    # determine the area (cgi-bin, nonssl) and adjust path relatively to this area
    my $isCgiDir = 0;
    if ($path =~ /^\/cgi-bin$/) { # /cgi-bin itself
      $isCgiDir = 1;
      $path = '/';
    } elsif ($path =~ /^\/cgi-bin\//) { # directories living under /cgi-bin
      $isCgiDir = 1;
      $path =~ s/^\/cgi-bin//;
    }
    my $pDirNode = XmlNode->new('pdir',
      'attributes' => {
        'name' => $path,
        'nonssl' => $isCgiDir ? 'false' : 'true',
        'cgi' => $isCgiDir ? 'true' : 'false'
      }
    );
    $pDirNode->setAttribute('title', $properties->{'title'}) if $properties->{'title'};
    my %usersMap = %{$properties->{'users'}} if (ref($properties->{'users'}) eq 'HASH');
    while (my ($username, $password) = each %usersMap) {
      $pDirNode->addChild(XmlNode->new('pduser',
        'attributes' => {
          'name' => $username
        },
        'children' => [
          CommonXmlNodes::encryptedPassword($password)
        ]
      ));
    } # for each user
    push @result, $pDirNode;
  } # for each protected directory

  return \@result;
}

sub _makePinfoNodes {
  my ($pinfoHash) = @_;

  my @result = map { 
    XmlNode->new('pinfo', 
      'attributes' => {'name' => $_},
      'content' => $pinfoHash->{$_}
    ) 
  } keys %$pinfoHash;

  return @result;
}

sub _makeDatabaseListNode {
  my ($client, $domain) = @_;

  my $dbAdminLogin = ConfixxConfig::getValue('mysqlUserUser');
  my $dbAdminPassword = ConfixxConfig::getValue('mysqlUserPw');

  my @databases = Dumper::getClientDatabaseNames($client);
  unless (@databases) {
    return;
  }

  return XmlNode->new('databases',
      'children' => [
          map { _makeDatabaseNode($dbAdminLogin, $dbAdminPassword, $_, $domain) } @databases 
      ]);
}

sub _makeDatabaseNode {
  my ($dbLogin, $dbPassword, $dbName, $domainName) = @_;
  my $dbType  = 'mysql';
  my $adminDb = 'mysql';

  my $dbNode = XmlNode->new('database',
      'attributes' => {
            'name'     => $dbName,
            'type'     => $dbType,
            'version'  => Db::MysqlUtils::getVersion()
      },
      'children' => [
            DatabaseContentDumper::getDatabaseContent($contentDumperBase, $dbLogin, $dbPassword, $dbName, $dbType, $domainName)
      ]
  );

  my $dbCon = Db::DbConnect::getDbConnect($dbType, $dbLogin, $dbPassword, $adminDb);
  # Skip 'root' user during DB users dump.
  Db::DbConnect::addDbUsers($dbNode, $dbName, $dbCon, ['root']);
  $dbCon->{'DISCONNECT'}->();

  return $dbNode;
}

sub _makeDomainMaillistsNode {
  my ($clientInfo_ptr) = @_;

  return XmlNode->new('maillists',
    'children' => [
      XmlNode->new('properties', 
        'children' => [ CommonXmlNodes::status($clientInfo_ptr->{'maxmaillist'} != 0, 'admin') ]
      )
    ]
  );
}

sub _makeSiteMaillistsNode {
  my ($clientInfo_ptr, $domainName) = @_;

  my $maillistInfos_ptr = Dumper::getMaillists($domainName);
  unless (@{$maillistInfos_ptr}) {
    return undef;
  }

  return XmlNode->new('maillists',
    'children' => [
      XmlNode->new('properties', 
        'children' => [ CommonXmlNodes::status($clientInfo_ptr->{'maxmaillist'} != 0, 'admin') ]
      ),
      map { _makeMaillistNode($_) } @{$maillistInfos_ptr}
    ]
  );
}

sub _makeMaillistNode {
  my ($maillistInfo_ptr) = @_;

  my $passwordNode = CommonXmlNodes::password('plain', $maillistInfo_ptr->{'pwd'}) if $maillistInfo_ptr->{'pwd'};

  my $maillistNode = XmlNode->new('maillist',
    'attributes' => {
      'name' => $maillistInfo_ptr->{'name'}
    },
    'children' => [
      CommonXmlNodes::status(! $maillistInfo_ptr->{'gesperrt'}, 'admin'),
      XmlNode->new('owner', 'content' => $maillistInfo_ptr->{'owner_mail'}),
      $passwordNode,
      map { XmlNode->new('recipient', 'content' => $_) } @{$maillistInfo_ptr->{'recipients'}}
    ]
  );
  
  return $maillistNode;
}

sub getTraffic {
  my ($client) = @_;

  my @result;

  # <doc>
  # It is still not clear if 'pop', 'other', 'geloescht' and 'log' fields from 'transfer' table
  # are still used in Confixx and what's their meaning, so we do not migrate them.
  # </doc>
  my %mapping = (
    'web' => 'http',
    # <doc>
    # There is no way to recognize if mail traffic was POP3/IMAP traffic or SMTP traffic,
    # so we put all Confixx mail traffic into Plesk POP3/IMAP traffic.
    # </doc>
    'email' => 'pop3-imap',
    'ftp' => 'ftp'
  );

  for my $trafficItem (Dumper::getTraffic($client)) {
    my ($day, $month, $year, $type, $direction, $amount);
    
    ($day, $month, $year) = ($trafficItem->{'day'}, $trafficItem->{'month'}, $trafficItem->{'year'});

    # <doc>
    # Confixx sumarizes traffic that is older than 3 months, we move summarized records to the last day of the month.
    # </doc>
    if ($day == 0) {
      $day = DateUtils::lastMonthDay($year, $month);
    }

    for my $confixxTrafficType (keys %mapping) {
      my $pleskTrafficType = $mapping{$confixxTrafficType};

      if ($trafficItem->{$confixxTrafficType} > 0) {
        push @result, {
          'day' => $day,
          'month' => $month,
          'year' => $year,
          'type' => $pleskTrafficType,
          # <doc>
          # There is no way to recognize direction of traffic in Confixx, so we always set it to 'in'.
          # </doc>
          'direction' => 'in',
          # In Confixx traffic is stored in kilobytes, convert to bytes
          'amount' => $trafficItem->{$confixxTrafficType} * 1024
        };
      }
    }
  }

  return @result;
}

sub _makeDnsSettingsNode {
  my $adminInfo  = Dumper::getAdminInfo();
  my @rawDnsRecords = map { s/##(\w+)##/<$1>/g; $_; } split "\n", $adminInfo->{'dnstemplate'};

  # add a Plesk standard NS template; weed out Confixx NS templates ##ns1## and ##ns2##
  push @rawDnsRecords, '<domain>. 86400 IN NS ns.<domain>';
  my @dnsRecords = Dumper::parseDnsRecords(grep { !/<ns[12]>/ } @rawDnsRecords);

  my $domainName = $adminInfo->{'hostname'};
  my $email      = "hostmaster.$domainName";
  my $status     = 1; # enabled

  return XmlNode->new('dns-settings', 
    'attributes' => {
        'recursion' => 'localnets' # Plesk default setting
    },
    'children' => [
        _makeDnsZoneNode($domainName, $email, $status, \@dnsRecords)
    ]
  );

}

1;
